open Printf
open MIPSOps
open Primitive
open PP
open Integer

(* ------------------------------------------------------------------------- *)

(* This exception is intentionally never caught. This allows us to obtain
   stack backtraces when debugging. *)

exception RuntimeError

(* ------------------------------------------------------------------------- *)

type value =
  | VInt of int32

let asInt (v : value) : int32 =
  match v with
  | VInt i ->
      i

(* ------------------------------------------------------------------------- *)

(* Interpreting primitive operations. We expect the number and nature
   of the parameters to be consistent; this is guaranteed by the
   typechecker. *)

let interpret_primitive (p : primitive) (actuals : value list) =
  match p, actuals with
  | Write, [ v ] ->
      fprintf stdout "%ld%!" (asInt v);
      None
  | Writeln, [ v ] ->
      fprintf stdout "%ld\n%!" (asInt v);
      None
  | Readln, [] ->
      let line = input_line stdin in
      begin try
	Some (VInt (Int32.of_string line))
      with Failure "int of string" ->
	fprintf stderr "readln: error: \"%s\" is not the representation of an integer.\n%!" line;
	raise RuntimeError
      end
  | _ ->
      assert false

(* ------------------------------------------------------------------------- *)

(* An environment maps an identifier (a name for a variable) to a
   modifiable value (a reference cell that holds a value). There is
   one environment for global variables and one for local
   variables. *)

type environment =
  value ref StringMap.t

(* ------------------------------------------------------------------------- *)

(* Another environment maps an identifier (a name for a procedure or
   function) to a definition. Procedure and function definitions are
   fixed, so no reference cells are involved here. *)

type definitions =
  PP.procedure StringMap.t

(* Interpreting function and procedure calls. This returns [None] for
   procedures and [Some result] for functions. *)

let rec interpret_call
  (defs : definitions)
  (genv : environment)
  (env : environment)
  (callee : callee)
  (actuals : expression list)
  : value option =

  (* Evaluate the actual arguments. [List.map] implements left to
     right evaluation order, and [PP] has left to right evaluation
     order, so this is consistent. *)

  let actuals = List.map (interpret_expression defs genv env) actuals in

  (* User-defined functions and primitive functions are treated
     differently. *)

  match callee with

  | CPrimitiveFunction p ->

      interpret_primitive p actuals

  | CUserFunction f ->

      failwith "UNIMPLEMENTED. FIX ME!"

(* ------------------------------------------------------------------------- *)

(* Interpreting expressions. *)

and interpret_expression
  (defs : definitions)
  (genv : environment)
  (env  : environment)
  (e    : expression)
  : value =
  
  match e with

  | EConst (ConstInt i) ->
      VInt i

  | _ ->
      failwith "UNIMPLEMENTED. FIX ME!"

(* ------------------------------------------------------------------------- *)

(* Interpreting conditions. This yields a Boolean result. *)

and interpret_condition
  (defs : definitions)
  (genv : environment)
  (env  : environment)
  (c    : condition)
  : bool =

  match c with

  | _ ->
      failwith "UNIMPLEMENTED. FIX ME!"

(* ------------------------------------------------------------------------- *)

(* Interpreting instructions. This yields no result, but execution can
   have side effects, that is, modify the value of some variables, or
   invoke primitive operations that perform input/output. *)

and interpret_instruction
  (defs : definitions)
  (genv : environment)
  (env  : environment)
  (i    : instruction)
  : unit =

  match i with

  | IProcCall (callee, es) ->
      begin match interpret_call defs genv env callee es with
      | None ->
	  ()
      | Some _ ->
	  assert false
      end


  | ISeq is ->
      List.iter (interpret_instruction defs genv env) is

  | _ ->
      failwith "UNIMPLEMENTED. FIX ME!"

(* ------------------------------------------------------------------------- *)

(* Interpreting programs. *)

let interpret (p : program) : unit =

  (* Create an environment that holds the global variables. *)

  let genv =
    StringMap.empty (* UNIMPLEMENTED. FIX ME! *)
  in

  (* Execute the program body. *)

  interpret_instruction p.defs genv StringMap.empty p.main

